﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System.Net.Sockets;

public class ChunkManager
{
    // chunks inside this radius will be actively loaded and visible
    private Vector3Int m_min_view_radius;

    // chunks within this radius stay active, but are not loaded actively. Chunks
    // outside this radius are set inactive
    private Vector3Int m_max_view_radius;

    // chunks within this radius are not deleted
    private Vector3Int m_cached_chunk_radius;

    private Transform m_reference_transform;

    private Vector3Int m_reference_chunk_position;

    private Vector3Int m_srqd_min_view_radius;
    private Vector3Int m_srqd_max_view_radius;
    private Vector3Int m_srqd_cached_chunk_radius;

    private WorldManager m_world_manager;

    private NetworkThread m_network_thread;

    private bool m_end_builder_threads = false;

    private Queue<MessageObject> m_received_message_queue;

    private Queue<ChunkData> m_create_chunk_data_queue;

    private Queue<Chunk> m_chunk_mesh_update_queue;

    // queue of chunk which are waiting for geometry to be applied
    private Queue<Chunk> m_chunk_geometry_build_queue;
    private Queue<ChunkData> m_chunk_data_to_cache_queue;

    private object m_handle = new object();

    private System.Threading.Thread[] m_threads = null;


    // chunks that are in our memory (either as active or inactive object in Unity)
    private Dictionary<Vector3Int, ChunkData> m_dict_cached_chunk_data;

    private Dictionary<Vector3Int, Chunk> m_dict_cached_chunks;

    // active (visible chunks)
    private Dictionary<Vector3Int, Chunk> m_dict_active_chunks;

    // chunks that are requested from server but not received yet
    private Dictionary<Vector3Int, bool> m_dict_chunk_requested;

    // blocked chunks. (e.g. geometry is build in another thread) They will not be removed
    private Dictionary<Vector3Int, bool> m_dict_chunk_blocked;

    private Dictionary<Vector3Int, bool> _m_dict_is_chunk_data_cached;

    private Dictionary<Vector3Int, bool> _m_dict_is_chunk_cached;

    private bool isChunkDataCached(Vector3Int p)
    {
        bool b;
        lock(_m_dict_is_chunk_data_cached)
        {
            b = _m_dict_is_chunk_data_cached.ContainsKey(p);
        }
        return b;
    }

    private void setChunkDataCached(Vector3Int p, bool b)
    {
        if (b)
        {
            lock(_m_dict_is_chunk_data_cached)
            {
                _m_dict_is_chunk_data_cached[p] = true;
            }
            return;
        }
        lock(_m_dict_is_chunk_data_cached)
        {
            _m_dict_is_chunk_data_cached.Remove(p);
        }
    }

    private void cacheChunkData(ChunkData cd)
    {
        m_dict_cached_chunk_data[cd.ChunkPositionVec] = cd;

        setChunkDataCached(cd.ChunkPositionVec, true);
    }

    private void uncacheChunkData(Vector3Int chunk_pos)
    {
        m_dict_cached_chunk_data.Remove(chunk_pos);

        setChunkDataCached(chunk_pos, false);
    }

    private bool isChunkCached(Vector3Int p)
    {
        bool b;
        lock(_m_dict_is_chunk_cached)
        {
            b = _m_dict_is_chunk_cached.ContainsKey(p);
        }
        return b;
    }

    private void setChunkCached(Vector3Int p, bool b)
    {
        if (b)
        {
            lock(_m_dict_is_chunk_cached)
            {
                _m_dict_is_chunk_cached[p] = true;
            }
            return;
        }
        lock(_m_dict_is_chunk_cached)
        {
            _m_dict_is_chunk_cached.Remove(p);
        }
    }

    private void cacheChunk(Chunk c)
    {
        m_dict_cached_chunks[c.ChunkPositionVec] = c;

        setChunkCached(c.ChunkPositionVec, true);
    }

    private void uncacheChunk(Vector3Int chunk_pos)
    {
        m_dict_cached_chunks.Remove(chunk_pos);

        setChunkCached(chunk_pos, false);
    }

    public bool end_builder_threads
    {
        get
        {
            bool v;
            lock(m_handle)
            {
                v = m_end_builder_threads;
            }
            return v;
        }
        set
        {
            lock(m_handle)
            {
                m_end_builder_threads = value;
            }
        }
    }

    public ChunkManager(Vector3Int min_view_radius,
                        Vector3Int max_view_radius,
                        Vector3Int cached_chunk_radius,
                        WorldManager world_manager,
                        NetworkThread network_thread,
                        Transform reference_transform,
                        int n_threads)
    {
        m_min_view_radius = min_view_radius;
        m_max_view_radius = max_view_radius;
        m_cached_chunk_radius = cached_chunk_radius;

        m_srqd_min_view_radius = Vector3Int.Scale(min_view_radius, min_view_radius);
        m_srqd_max_view_radius = Vector3Int.Scale(max_view_radius, max_view_radius);
        m_srqd_cached_chunk_radius = Vector3Int.Scale(cached_chunk_radius, cached_chunk_radius);

        m_world_manager = world_manager;
        m_network_thread = network_thread;
        m_reference_transform = reference_transform;

        m_dict_active_chunks = new Dictionary<Vector3Int, Chunk>();
        m_dict_cached_chunks = new Dictionary<Vector3Int, Chunk>();
        m_dict_cached_chunk_data = new Dictionary<Vector3Int, ChunkData>();
        m_dict_chunk_requested = new Dictionary<Vector3Int, bool>();
        m_dict_chunk_blocked = new Dictionary<Vector3Int, bool>();
        _m_dict_is_chunk_data_cached = new Dictionary<Vector3Int, bool>();
        _m_dict_is_chunk_cached = new Dictionary<Vector3Int, bool>();


        m_received_message_queue = new Queue<MessageObject>();
        m_chunk_mesh_update_queue = new Queue<Chunk>();
        m_chunk_geometry_build_queue = new Queue<Chunk>();
        m_create_chunk_data_queue = new Queue<ChunkData>();
        m_chunk_data_to_cache_queue = new Queue<ChunkData>();

        m_threads = new System.Threading.Thread[n_threads];

        for (int i = 0; i < n_threads; i++)
        {
            // Start chunk builder thread
            m_threads[i] = new System.Threading.Thread(RunChunkBuilderThread);
            m_threads[i].Start();
        }
        
    }

    public void processServerMessage(MessageObject mo)
    {
        //TODO: check datatype
        Vector3Int chunk_pos = new Vector3Int(mo.position[2], mo.position[1], mo.position[0]);

        if (m_dict_chunk_requested.ContainsKey(chunk_pos))
        {

            Vector3Int tmp = chunk_pos - m_reference_chunk_position;
            tmp = Vector3Int.Scale(tmp,tmp);

            if (!InRange(tmp, m_srqd_max_view_radius))
            {
                Debug.LogWarning("Lost Chunk: " + chunk_pos.ToString());
                return;
            }

            ChunkData cd = new ChunkData(mo);
            lock(m_create_chunk_data_queue)
            {
                m_create_chunk_data_queue.Enqueue(cd);
            }

        }
        else
        {
            Debug.LogWarning("Received chunk we not asked for: " + mo.position.ToString());
        }
    }

    private void AppendNewChunkToBuild(ChunkData cd)
    {
        Debug.Assert(cd.IsFullyConnected);

        Chunk c = m_world_manager.InstantiateNewChunk(cd);

        m_dict_chunk_blocked[cd.ChunkPositionVec] = true;

        lock(m_handle)
        {
            m_chunk_geometry_build_queue.Enqueue(c);
        }
    }
    

    private void RunChunkBuilderThread()
    {
        int count_data;
        int count_geometry;
        Chunk c = null;;
        ChunkData cd = null;
        bool success;

        while (!end_builder_threads)
        {
            success = false; 

            lock(m_handle)
            {
                count_geometry = m_chunk_geometry_build_queue.Count;
            }
            lock (m_create_chunk_data_queue)
            {
                count_data = m_create_chunk_data_queue.Count;
            }

            if(count_geometry == 0 && count_data == 0)
            {
                System.Threading.Thread.Sleep(50);
                continue;
            }
            
            if (count_data > count_geometry)
            {
                lock (m_create_chunk_data_queue)
                {
                    if (m_create_chunk_data_queue.Count > 0)
                    {
                        cd = m_create_chunk_data_queue.Dequeue();
                        success = true;
                    }
                }

                if (!success)
                {
                    // another thread was faster
                    continue;
                }

                cd.analyseData();


                lock(m_chunk_data_to_cache_queue)
                {
                    m_chunk_data_to_cache_queue.Enqueue(cd);
                }
            }
            else
            {
                // looking for new chunk geometry to build
                

                lock(m_handle)
                {
                    if (m_chunk_geometry_build_queue.Count > 0)
                    {
                        c = m_chunk_geometry_build_queue.Dequeue();
                        success = true;
                    }
                }

                if (!success)
                {
                    //another thread was faster
                    continue;
                }

                c.BuildInnerMesh();

                Vector3Int chunk_pos = c.ChunkPositionVec;

                // TODO: connect neighbors and build border regions
                for (int i = 0; i < 6; i++)
                {
                    c.BuildBorderGeometry((ChunkData.Direction)i);
                }

                c.ConcatenateGeometry();


                lock(m_handle)
                {
                    m_chunk_mesh_update_queue.Enqueue(c);
                }


            }

            
        }
        Debug.Log("Closing Chunk Builder Thread");
    }


    private bool isChunkActive(Vector3Int chunk_pos)
    {
        return m_dict_active_chunks.ContainsKey(chunk_pos);
    }

    public void requestChunkFromServer(Vector3Int chunk_pos)
    {
    
        if (!m_dict_chunk_requested.ContainsKey(chunk_pos) && !isChunkDataCached(chunk_pos))
        {
            m_dict_chunk_requested[chunk_pos] = true;
            //request chunk
            // build a chunk request to server:
            MessageObject mo = new MessageObject();
            mo.datatype = "request_chunk";

            // TODO: reverse z and x since unity uses a left handed coordinate system
            mo.position = new int[] {chunk_pos.z, chunk_pos.y, chunk_pos.x};
            m_network_thread.SendMessage(mo);
        }
    }

    public void ApplyUpdates(int max_chunk_updates=1)
    {
        int i = max_chunk_updates;
        int n;
        Chunk c;
        lock(m_handle)
        {
            n = m_chunk_mesh_update_queue.Count;
        }
        while(n > 0 && i > 0)
        {
            lock(m_handle)
            {
                c = m_chunk_mesh_update_queue.Dequeue();
            }

            c.AssignMesh();

            i -= 1;
            n -= 1;

            m_dict_chunk_blocked.Remove(c.ChunkPositionVec);

            cacheChunk(c);
            
            m_dict_active_chunks[c.ChunkPositionVec] = c;

        }
    }



    private bool InRange(Vector3Int sqrd_dist, Vector3Int sqrd_range)
    {

        Vector3Int tmp =  sqrd_range - sqrd_dist;

        return tmp.x > 0 && tmp.y > 0 && tmp.z > 0;
    }

    public void UpdateViewRange()
    {
        //getting new Chunk Data:
        int cache_count;
        lock(m_chunk_data_to_cache_queue)
        {
            cache_count = m_chunk_data_to_cache_queue.Count;
            
        }

        while (cache_count > 0)
        {
            ChunkData cd;
            lock (m_chunk_data_to_cache_queue)
            {
                cd = m_chunk_data_to_cache_queue.Dequeue();
                
            }

            // check for neighbors TODO: make this in builder thread!
            for (int i = 0; i < 6; i++)
            {
                Vector3Int neighbor_pos = cd.ChunkPositionVec + ChunkData.DirectionNormalsInt[i];
                if (isChunkDataCached(neighbor_pos))
                {
                    ChunkData neighbor = m_dict_cached_chunk_data[neighbor_pos];
                    cd.ConnectChunkData(neighbor, (ChunkData.Direction)i);
                    neighbor.ConnectChunkData(cd, (ChunkData.Direction) ((i+3)%6));

                }
            }
            cacheChunkData(cd);
            m_dict_chunk_requested.Remove(cd.ChunkPositionVec);
            cache_count --;
        }

        Vector3Int current_chunk_position = Vector3Int.RoundToInt(m_reference_transform.position);
		current_chunk_position.x /= ChunkData.N_CHUNK_SIDELENGTH;
		current_chunk_position.y /= ChunkData.N_CHUNK_SIDELENGTH;
		current_chunk_position.z /= ChunkData.N_CHUNK_SIDELENGTH;

        m_reference_chunk_position = current_chunk_position;

        Stack<Vector3Int> to_remove = new Stack<Vector3Int>();

        // check for chunks to remove:
        foreach(KeyValuePair<Vector3Int, ChunkData> entry in m_dict_cached_chunk_data)
        {
            // do something with entry.Value or entry.Key
            Vector3Int tmp = entry.Key - m_reference_chunk_position;
            tmp = Vector3Int.Scale(tmp,tmp);

            if (m_dict_chunk_blocked.ContainsKey(entry.Key))
            {
                continue;
            }

            if (m_dict_active_chunks.ContainsKey(entry.Key) && !InRange(tmp, m_srqd_max_view_radius))
            {
                Debug.Assert(entry.Value.ParentChunk != null);
                if (entry.Value.ParentChunk != null)
                    entry.Value.ParentChunk.gameObject.SetActive(false);
                else
                    Debug.LogError("corrupted chunk: " + entry.Key.ToString());
                m_dict_active_chunks.Remove(entry.Key);
            }


            if (!InRange(tmp, m_srqd_cached_chunk_radius))
            {
                to_remove.Push(entry.Key);
            }

        }

        while (to_remove.Count > 0)
        {
            Vector3Int k = to_remove.Pop();
            uncacheChunkData(k);

            if (m_dict_cached_chunks.ContainsKey(k))
            {
                Chunk c = m_dict_cached_chunks[k];
                uncacheChunk(k);
                UnityEngine.Object.Destroy(c);
            }
        }

        // reload view range:
        for (int x=-m_min_view_radius.x; x < m_min_view_radius.x; x++)
        {
            for (int y=-m_min_view_radius.y; y < m_min_view_radius.y; y++)
            {
                for (int z=-m_min_view_radius.z; z < m_min_view_radius.z; z++)
                {
                    Vector3Int p = new Vector3Int(x, y, z) + m_reference_chunk_position;
                    if (m_dict_active_chunks.ContainsKey(p) 
                        || m_dict_chunk_requested.ContainsKey(p) 
                        || m_dict_chunk_blocked.ContainsKey(p))
                    {
                        // TODO: the if above is a hell of a mess!
                        continue;
                    }

                    
                    
                    if (isChunkDataCached(p))
                    {
                        if (!isChunkCached(p))
                        {
                            
                            ChunkData cd = m_dict_cached_chunk_data[p];
                            
                            if (cd.IsFullyConnected )
                            {
                                if (cd.isBuiltNecessary)
                                {
                                    AppendNewChunkToBuild(cd);
                                }
                            }

                        }
                        else
                        {
                            Chunk c;
                        
                            c = m_dict_cached_chunks[p];
                            
                            c.gameObject.SetActive(true);
                            m_dict_active_chunks[p] = c;
                        }
                    }
                    
                    else if (!m_dict_chunk_requested.ContainsKey(p))
                    {
                        requestChunkFromServer(p);
                    }
                }
            }
        }

    }

}
